@@ -5,46 +5,53 @@ package App::ccdiff;
 use 5.014000;
 use warnings;
 use charnames ();
-use List::Util      qw( first  max     );
-use Term::ANSIColor qw(:constants color);
-use Getopt::Long    qw(:config bundling);
+use File::Find;
+use Encode          qw( encode                      );
+use List::Util      qw( first  max                  );
+use Term::ANSIColor qw(:constants color             );
+use Getopt::Long    qw(:config bundling noignorecase);
 
-our $VERSION = "0.32";
+our $VERSION = "0.35";
 our $CMD     = $0 =~ s{.*/}{}r;
 
 sub usage {
     my $err = shift and select STDERR;
     say "usage: $CMD [options] file1 [file2]";
+    say "       $CMD [options] dir1   dir2";
     say "       $CMD --man | --info";
     say "	file1 or file2 can be - (but not both)";
-    say "   -V    --version      Show version and exit";
-    say "   -v[1] --verbose[=1]  Set verbosity";
+    say "   -V     --version      Show version and exit";
+    say "   -v[1]  --verbose[=1]  Set verbosity";
     say "  Diff options:";
-    say "   -U    --utf-8                 Input is in UTF-8";
-    say "   -u[3] --unified=3             Show a unified diff";
-    say "         --no-header             Skip header with file names/stamps";
-    say "   -I    --index                 Add indices to the change chunks";
-    say "   -I n  --index=4               Only show chunk n";
-    say "   -w    --ignore-all-space      Ignore all whitespace";
-    say "   -b    --ignore-space-change   Ignore horizontal whitespace changes";
-    say "   -Z    --ignore-trailing-space Ignore whitespace at line ending";
-    say "   -B    --ignore-blank-lines    Ignore changes where lines are all blank";
-    say "   -i    --ignore-case           Ignore case changes";
+    say "   -U     --utf-8                 Input is in UTF-8";
+    say "   -u[3]  --unified=3             Show a unified diff";
+    say "          --no-header             Skip header with file names/stamps";
+    say "   -I     --index                 Add indices to the change chunks";
+    say "   -I n   --index=4               Only show chunk n";
+    say "   -w     --ignore-all-space      Ignore all whitespace";
+    say "   -b     --ignore-space-change   Ignore horizontal whitespace changes";
+    say "   -Z     --ignore-trailing-space Ignore whitespace at line ending";
+    say "   -B     --ignore-blank-lines    Ignore changes where lines are all blank";
+    say "   -i     --ignore-case           Ignore case changes";
+    say "   --dc=C --diff-class=C          Select Algorithm::Diff (AD or PP)";
+    say "                                      or Algorithm::Diff::XS (XS = default)";
     say "  Other options:";
-    say "   -t n  --threshold=2  Horizontal line diff threshold";
-    say "   -h n  --heuristics=n Horizontal char diff threshold";
-    say "   -e n  --ellipsis=n   Compress horizontal equal sections";
-    say "   -m    --markers      Use markers to indicate change positions";
-    say "   -a    --ascii        Use ASCII instead of Unicode indicators";
-    say "         --no-color     Reset all colors to none.";
-    say "         --list-colors  List available colors and exit";
-    say "         --old=red      Color to indicate removed content";
-    say "         --new=green    Color to indicate added   content";
-    say "         --bg=white     Background color for colored indicators";
-    say "   -p    --pink         Shortcut for --old=magenta";
-    say "   -r    --reverse      Reverse/invert the colors of the indicators";
-    say "   -s    --swap         Swap old/new color indicators";
-    say "         --settings     Show default settings (after reading rc)";
+    say "   -t n   --threshold=2  Horizontal line diff threshold";
+    say "   -h n   --heuristics=n Horizontal char diff threshold";
+    say "   -e n   --ellipsis=n   Compress horizontal equal sections";
+    say "   -m     --markers      Use markers to indicate change positions";
+    say "   -a     --ascii        Use ASCII instead of Unicode indicators";
+    say "          --no-color     Reset all colors to none.";
+    say "          --list-colors  List available colors and exit";
+    say "          --old=red      Color to indicate removed content";
+    say "          --new=green    Color to indicate added   content";
+    say "          --bg=white     Background color for colored indicators";
+    say "          --old-label=L  Set file/stream indicator for header old";
+    say "          --new-label=L  Set file/stream indicator for header new";
+    say "   -p     --pink         Shortcut for --old=magenta";
+    say "   -r     --reverse      Reverse/invert the colors of the indicators";
+    say "   -s     --swap         Swap old/new color indicators";
+    say "          --settings     Show default settings (after reading rc)";
     exit $err;
     } # usage
 
@@ -63,6 +70,9 @@ my %rc = (
     header	=> 1,	# A color name is allowed
     heuristics	=> 0,
     index	=> 0,
+    iwbzusepp	=> 0,
+    new_label	=> undef,
+    old_label	=> undef,
     markers	=> 0,
     new		=> "green",
     old		=> "red",
@@ -99,7 +109,10 @@ my $new_color = $rc{new};
 my $rev_color = $rc{bg};
 my $cli_color = $ENV{CLICOLOR};	# https://bixense.com/clicolors/
 my $no_colors = $ENV{NO_COLOR}; # https://no-color.org
+my $old_label = $rc{old_label};
+my $new_label = $rc{new_label};
 my $list_colors;
+my $diff_class;
 
 if ($no_colors) {
     $ENV{CLICOLOR_FORCE}	and $no_colors = 0;
@@ -118,6 +131,8 @@ unless (caller) {
 	  "info"	=> sub { pod_text  (); },
 
 	"U|utf-8!"		=> \$opt_U,
+	  "dc|diff-class=s"	=> \$diff_class,
+	  "pp!"			=> sub { $diff_class = "PP" },
 
     #   "c|context:3"		=> \$opt_c,	# implement context-diff?
 	"u|unified:3"		=> \$opt_u,
@@ -141,6 +156,9 @@ unless (caller) {
 	"E|ignore-tab-expansion!"		=> \$opt_E, # NYI
 	"B|ignore-blank-lines!"			=> \$opt_B, # Partly implemented
 
+	  "old-label|label-old=s"		=> \$old_label,
+	  "new-label|label-new=s"		=> \$new_label,
+
 	"p|pink!"		=> sub { $old_color = "magenta" },
 	  "old=s"		=> \$old_color,
 	  "new=s"		=> \$new_color,
@@ -230,8 +248,12 @@ sub set_options {
     $clr_new = $opt_r ? $clr_rev . $bg_new : $clr_grn . $bg_new;
     $opt_s and ($clr_new, $clr_old) = ($clr_old, $clr_new);
     # Indicators
-    $opt_a and
+    if ($opt_a) {
 	@rc{qw( chr_old chr_new chr_cml chr_cmr chr_eli chr_eli_v )} = qw( ^ ^ > < - <> );
+	}
+    elsif (!$opt_U) {
+	$rc{$_} = encode ("utf-8", $rc{$_}) for qw( chr_old chr_new chr_cml chr_cmr chr_eli chr_eli_v );
+	}
     $chr_old = $clr_old . $rc{chr_old} . $reset;
     $chr_new = $clr_new . $rc{chr_new} . $reset;
     $chr_cml = $clr_dbg . $rc{chr_cml} . $reset;
@@ -243,20 +265,73 @@ sub set_options {
     $chr_eli = $opt_v >= 2 ? $rc{chr_eli_v} : $rc{chr_eli};
     $opt_m && $opt_v > 1 && length ($chr_eli) > 1 and $opt_m = 0;
 
-    $cmp_sub = $opt_i || $opt_b || $opt_Z ? { keyGen => sub {
-	my $line = shift;
-	$opt_i and $line = lc $line;
-	$opt_Z and $line =~ s/[ \t]+$//g;
-	$opt_b and $line =~ s/[ \t]+/ /g;
-	return $line;
-	}} : undef;
+    if ($opt_i || $opt_b || $opt_Z) {
+	$rc{iwbzusepp} and $diff_class = "Algorithm::Diff";
+	$cmp_sub = {
+	    keyGen => sub {
+		my $line = shift;
+		$opt_i and $line = lc $line;
+		$opt_Z and $line =~ s/[ \t\r]+$//g;
+		$opt_b and $line =~ s/[ \t]+/ /g;
+		return $line;
+		}
+	    };
+	}
     } # set_options
 
-my $diff_class = eval { require Algorithm::Diff::XS; "Algorithm::Diff::XS"; }
-	      || eval { require Algorithm::Diff;     "Algorithm::Diff";     }
-	      or die "Cannot load Algorithm::Diff:\n";
+if ($diff_class) {
+    if ($diff_class =~ m/^(?:ad|pp|algorithm(?:-|::)diff(?:(?:-|::)pp)?)$/i) {
+	$diff_class =
+	    eval { require Algorithm::Diff;     "Algorithm::Diff";     }
+	}
+    elsif ($diff_class =~ m/^(?:adx|xs|algorithm(?:-|::)diff(?:-|::)xs)$/i) {
+	$diff_class = "Algorithm::Diff::XS";
+	$diff_class =
+	    eval { require Algorithm::Diff::XS; "Algorithm::Diff::XS"; }
+	}
+    else {
+	die "$diff_class is an unsupported Diff class\n";
+	}
+    }
+else {
+    $diff_class =
+	eval { require Algorithm::Diff::XS; "Algorithm::Diff::XS"; }
+     || eval { require Algorithm::Diff;     "Algorithm::Diff";     };
+    }
+$diff_class or die "Cannot load Algorithm::Diff:\n";
 
-caller or ccdiff (@ARGV);
+unless (caller) {
+    if (@ARGV == 2 && -d $ARGV[0] && -d $ARGV[1]) {
+	my ($d_from, $d_to) = @ARGV;
+	my %fn;
+	my @f;
+	foreach my $idx (0, 1) {
+	    my $dir = $ARGV[$idx];
+	    find (sub {
+		-f && -s or return;
+		my $fn = $File::Find::name =~ s{^$dir/*}{}r;
+		$fn{$fn}++;
+		$f[$idx]{$fn}++;
+		}, $dir);
+	    }
+	foreach my $fn (sort keys %fn) {
+	    if ($f[0]{$fn} && $f[1]{$fn}) {
+		ccdiff ("$d_from/$fn", "$d_to/$fn");
+		}
+	    elsif ($f[0]{$fn}) {
+		say "Only in $d_from: $fn";
+		}
+	    else {
+		say "Only in $d_to: $fn";
+		}
+	    delete $f[0]{$fn};
+	    }
+	}
+    else {
+	ccdiff (@ARGV);
+	}
+    exit 0;
+    }
 
 sub ccdiff {
     my $f1 = shift or usage (1);
@@ -265,7 +340,7 @@ sub ccdiff {
     -b $f1 || -c $f1 || -b $f2 || -c $f2 and
 	die "Character and block devices are not supported\n";
     -d $f1 || -d $f2 and
-	die "$CMD does not yet support recursive diff and/or directory diff\n";
+	die "$CMD does not support directory diff\n";
 
     my $fh;
 
@@ -291,6 +366,8 @@ sub ccdiff {
 	    $o eq "markers"			and $opt_m = $v;
 	    $o eq "new"				and $new_color = $v;
 	    $o eq "old"				and $old_color = $v;
+	    $o eq "new-label"			and $new_label = $v;
+	    $o eq "old-label"			and $old_label = $v;
 	    $o eq "reverse"			and $opt_r = $v;
 	    $o eq "swap"			and $opt_s = $v;
 	    $o eq "threshold"			and $opt_t = $v;
@@ -311,10 +388,12 @@ sub ccdiff {
 
     $f1 eq "-" && $f2 eq "-" and usage (1);
 
-               binmode STDERR, ":encoding(utf-8)";
+        binmode STDERR, ":encoding(utf-8)";
 
-    $opt_U and binmode STDIN,  ":encoding(utf-8)";
-    $opt_U and binmode STDOUT, ":encoding(utf-8)";
+    if ($opt_U) {
+	binmode STDIN,  ":encoding(utf-8)";
+	binmode STDOUT, ":encoding(utf-8)";
+	}
 
     my @d1 = ref $f1 eq "ARRAY" ? @$f1 : $f1 eq "-" ? <STDIN> : do {
 	open my $fh, "<", $f1 or die "$f1: $!\n";
@@ -335,13 +414,15 @@ sub ccdiff {
 	    }
 	my $nl = max length $f1, length $f2, 7;
 	my $sl = $hc ? ($ENV{COLUMNS} || 80) - 4 - $nl : 1;
-	my @h = map { -f $_
+	my $hi = -1;
+	my @hi = ($old_label, $new_label);
+	my @h  = map { $hi++; -f $_
 	    ? { tag   => "",
 		name  => $_,
 		stamp => scalar localtime ((stat $_)[9]),
 		}
 	    : { tag   => "",
-		name  => "*STDIN",
+		name  => $hi[$hi] // "*STDIN",
 		stamp => scalar localtime,
 		}
 	    } $f1, $f2;
@@ -594,6 +675,7 @@ ccdiff - Colored Character diff
 =head1 SYNOPSIS
 
  ccdiff [options] file1|- file2|-
+ ccdiff [options] dir1    dir2
 
  ccdiff --help
  ccdiff --man
@@ -601,6 +683,14 @@ ccdiff - Colored Character diff
 
 =head1 DESCRIPTION
 
+Show the diff between two files on a character by character base. In contrast to
+the standard diff tools, this tool uses the diff algorithm horizontally for each
+line in the vertical diff, highlighting the changes. This is very handy in hard
+to spot changes like C<O> to C<0>, C<I> to C<l> or C<1> and whitespace.
+
+If there are two argument, and both are a folder/directory, a recursive diff is
+executed. This is not available whan used as a (sub)class.
+
 =head1 OPTIONS
 
 =head2 Command line options
@@ -627,6 +717,30 @@ Show this manual using pod2text.
 
 All I/O (streams to compare and standard out) are in UTF-8.
 
+=item --diff-class=C --dc=C --pp
+
+Select the class used to execute the diff. By default C<ccdiff> will select
+the first available out of C<Algorithm::Diff::XS> or C<Algorithm::Diff>.
+
+Sometime the C<XS> version fails on encoding and the pure-perl version will
+work just fine. You can force C<ccdiff> to use either
+
+Select the pure-perl version with any of C<PP>, C<AD>, C<Algorthm::Diff>,
+C<Algorithm-Diff>, or C<Algorithm::Diff::PP> (case insensitive).
+For convenience, C<--dc=pp> can be abbreviated to C<--pp>.
+
+ --pp
+ --dc=pp
+ --dc=algorithm-diff
+ --diff-class=Algorithm::Diff::PP
+
+Select the XS version with any of C<XS>, C<ADX>, C<Algorthm::Diff::XS>, or
+C<Algorithm-Diff-XS> (case insensitive).
+
+ --dc=xs
+ --dc=algorithm-diff-xs
+ --diff-class=Algorithm::Diff::XS
+
 =item --unified[=3] -u [3]
 
 Generate a unified diff. The number of context lines is optional. When omitted
@@ -1019,6 +1133,27 @@ If the value is a valid supported color, it will show the header in that
 color scheme.  To disable the header set it to C<0> in the RC file or use
 C<--no-header> as a command line argument.
 
+=item old-label
+
+=item new-label
+
+Defines the tag(s) in the header for the source file(s)/stream(s).
+
+ App::ccdiff::ccdiff ($left, $right);
+
+ =>
+
+ < *STDIN  Fri Nov  7 10:14:51 2025
+ > *STDIN  Fri Nov  7 10:14:51 2025
+
+ App::ccdiff::ccdiff ($left, $right,
+    { "old-label" => "OLD", "new-label" => "NEW" });
+
+ =>
+
+ < OLD     Fri Nov  7 10:14:51 2025
+ > NEW     Fri Nov  7 10:14:51 2025
+
 =item verbose
 
  verbose : cyan
@@ -1135,7 +1270,7 @@ in a line when verbose level is 3 and up.
 
 =item chr_eli
 
- chr_eli : U+21B0
+ chr_eli : U+2508
 
 Defines the character used to indicate omitted text in large unchanged text
 when C<--ellipsis>/C<-e> is in effect.
@@ -1176,6 +1311,13 @@ length of the compressed part is put between the characters.
 
 A suggested alternative might be U+21E4U+21E5
 
+=item iwbZusePP
+
+As L<Algorithm::Diff::XS> is fast but is not 100% drop-in compliant with
+L<Algorithm::Diff> and options C<-i>, C<-w>, C<-b>, and C<-Z> are likely to
+not work in the C<XS> version, this option allows automatic switching to
+the slower version if any of these options is selected.
+
 =back
 
 =head1 Git integration
@@ -1212,6 +1354,11 @@ compared to e.g. GNU diff. Installing L<Algorithm::Diff::XS> will make
 C<ccdiff> a lot faster. C<ccdiff> will choose L<Algorithm::Diff::XS> if
 available.
 
+Note however that options like C<-i>, C<-w>, C<-b>, and C<-Z> are likely to
+be a no-op in L<Algorithm::Diff::XS>, as that has not been implemented and
+it is rather unlikely it will be. Choose C<--dc=pp> then. If you want this
+switch always, set C<iwbZusePP = 1> in one of your L</Configuration files>.
+
 =head1 SEE ALSO
 
 L<Algorithm::Diff::XS>, L<Algorithm::Diff>, L<Text::Diff>
@@ -1222,7 +1369,7 @@ H.Merijn Brand
 
 =head1 COPYRIGHT AND LICENSE
 
- Copyright (C) 2018-2023 H.Merijn Brand.  All rights reserved.
+ Copyright (C) 2018-2026 H.Merijn Brand.  All rights reserved.
 
 This library is free software;  you can redistribute and/or modify it under
 the same terms as The Artistic License 2.0.
